#include "CSunburstCluster.H"
#include "XmdvTool.H"
#include "stdio.h"
#include <math.h>
#include <string.h>

static double alpha, min_angle;

CSunburstCluster::CSunburstCluster()
{
	children = NULL;
	siblings = NULL;
    selected = false;
    show_children = true;
    current = false;
	radius = 0;
}


CSunburstCluster::~CSunburstCluster()
{
    if ( children != NULL )
        delete children;
    if ( siblings != NULL )
        delete siblings;
}


//reset level of this node and its children and next silbings, find the maximum level among them;
//current_level: level to be signed to this node; max_level: will be the maximum level of itself, its children and next silbings 
void CSunburstCluster::ResetLevels(int current_level, int &max_level)
{
	level = current_level;
	
	if ( level > max_level )
		max_level = level;

	if ( children != NULL && show_children == true)
		children->ResetLevels(current_level+1, max_level);
	if ( siblings != NULL )
		siblings->ResetLevels(current_level, max_level);
}


void CSunburstCluster::ResetEntries()
{
	if ( children == NULL )
    {
        entries = 1;
        return;
    }

    int count = 0;
    CSunburstCluster *child = children;
    while ( child != NULL )
    {
        child->ResetEntries();
        count += child->entries;
        child = child->siblings;
    }
    entries = count;

    if ( siblings != NULL )
		siblings->ResetEntries();
}

void CSunburstCluster::ResetCurrentList(CLinkList* current_list)
{
	if ( current )
		current_list->AddNodeAtEnd(this);

	if ( siblings != NULL )
		siblings->ResetCurrentList(current_list);

	if ( children != NULL )
		children->ResetCurrentList(current_list);
}


//reset the start_angle of this node;
//start_angle: the new start angle to be asigned to this node; unit_angle: angle responds to one dimension;
void CSunburstCluster::ResetAngles(double start_angle, double unit_angle)
{
	this->start_angle = start_angle;
	this->sweep_angle = unit_angle * entries;

	if ( children != NULL )
		children->ResetAngles(start_angle, unit_angle);
	if ( siblings != NULL )
		siblings->ResetAngles(GetEndAngle(), unit_angle);
}


double CSunburstCluster::GetEndAngle()
{
	double angle = start_angle + sweep_angle;
	while ( angle >= 360 )
	{
		angle = angle - 360;
	}
	return angle;
}


double CSunburstCluster::GetMidAngle()
{
	double angle = start_angle + sweep_angle/2.;
	if ( angle >= 360 )
		angle = angle - 360;
	return angle;
}

/*
//reset color of this node; color responds to its mid angle. 
void CSunburstCluster::ResetColors(colormap CMAP)
{
	double mid_angle = GetMidAngle();
	double color_idx_test_punit = mid_angle * COLORCOEF;
    int color_idx = (int) color_idx_test_punit;
	color = CMAP[color_idx];

	if ( children != NULL )
		children->ResetColors(CMAP);
	if ( siblings != NULL )
		siblings->ResetColors(CMAP);
}
*/


void CSunburstCluster::ResetColors(colormap CMAP)
{
	if ( children == NULL )
	{
		double mid_angle = GetMidAngle();
		double color_idx_test_punit = mid_angle * COLORCOEF;
//		int color_idx = (int)mid_angle * COLORCOEF;
        int color_idx = (int) color_idx_test_punit;
		color = CMAP[color_idx];
		return;
	}

	double r, g, b;
	r = 0;
	g = 0;
	b = 0;

	CSunburstCluster* child = children;
	while ( child != NULL )
	{
		child->ResetColors(CMAP);
		r += child->color.R * child->entries;
		g += child->color.G * child->entries;
		b += child->color.B * child->entries;
		child = child->siblings;
	}

	r /= entries;
	g /= entries;
	b /= entries;

	color.R = r;
	color.G = g;
	color.B = b;

	return;
}


//add a child to this node; the new child will be at the beginning of the child link list.
//child: the new child to be added;
void CSunburstCluster::AddChild(CSunburstCluster *child)
{
    child->siblings = children;
    child->parent = this;
    child->level = level+1;
    children = child;
}

//add a silbing to this node; the new sibling will be its next sibling;
//sibling: the new silbing to be added;
void CSunburstCluster::AddSibling(CSunburstCluster *sibling)
{
    sibling->parent = parent;
    sibling->level = level;
    sibling->siblings = siblings;
    siblings = sibling;
}

//remove a child of this node;
//rm_child: the child to be removed;
void CSunburstCluster::RemoveChild(CSunburstCluster *rm_child)
{
    CSunburstCluster* child = children;
	CSunburstCluster* next_child;
    if ( child == rm_child )
    {
         children = child->siblings;
    }
    else
    {
        while ( child != NULL )
        {
            next_child = child->siblings;
            if ( next_child == rm_child )
            {
                 child->siblings = next_child->siblings;
                 rm_child->siblings = NULL;
                 rm_child->parent = NULL;
                 child = NULL;
            }
            else
            {
                child = next_child;
            }
        }
    }
    return;
}


//distortion
//return the distance from start_angle to end_angle
//  start_angle            end_angle
//       | --  distance   ------|
double CSunburstCluster::GetAnglesDistance(double start_angle, double end_angle)
{
	if ( end_angle < start_angle )
		end_angle += 360;
	return end_angle - start_angle;
}

void CSunburstCluster::ChangeSweepAngle()
{
	if ( alpha > 1. )
		IncreaseSweepAngle();
	else 
		DecreaseSweepAngle();
}


//Increase sweep_angle by alpha, alpha > 0
void CSunburstCluster::IncreaseSweepAngle()
{
	sweep_angle *= alpha;

	CSunburstCluster* child = children;
	while ( child != NULL )
	{
		child->IncreaseSweepAngle();
		child = child->siblings;
	}
	return;
}


//decrease sweep_angle by alpha, alpha < 0, sweep_angle should > entries*min_angle
void CSunburstCluster::DecreaseSweepAngle()
{
    if ( children == NULL )
	{
		sweep_angle *= alpha;
		if ( sweep_angle < entries*min_angle )
			sweep_angle = entries*min_angle;
		return;
	}

	CSunburstCluster* child = children;
    double new_sweep_angle = 0;
	while ( child != NULL )
	{
		child->DecreaseSweepAngle();
		new_sweep_angle += child->sweep_angle;
		child = child->siblings;
	}

	sweep_angle = new_sweep_angle;
	return;
}


//change start angle of this and all its children.
void CSunburstCluster::ChangeStartAngle(double new_start_angle)
{
    start_angle = new_start_angle;

    CSunburstCluster* child = children;
	double angle = start_angle;
    while ( child != NULL )
    {
        child->ChangeStartAngle( angle );
		angle = child->GetEndAngle();
		child = child->siblings;
    }
}


//return double between 0 and 360; 
double CSunburstCluster::StandarizeAngle(double angle)
{
	while ( angle < 0 )
	{
		angle += 360;
		return angle;
	}

	while ( angle > 360 )
	{
		angle -= 360;
		return angle;
	}

    return angle;
}

CSunburstCluster* CSunburstCluster::FindLastSibling()
{
	CSunburstCluster *sibling = this; 
	while ( sibling->siblings != NULL )
	{
		sibling = sibling->siblings;
	}
	return sibling;
}


CSunburstCluster* CSunburstCluster::FindBeforeSibling()
{
	CSunburstCluster *sibling = this->parent->children; 
	while ( siblings != NULL && sibling->siblings != this )
	{
		sibling = sibling->siblings;
	}
	return sibling;
}

void CSunburstCluster::ChangeSweepAngle(bool toward_left, double new_angle, double minimum_angle)
{
	if ( toward_left )
	{
		double delta_angle = start_angle - new_angle;
		if ( delta_angle < -180 )
			delta_angle += 360;
		else if ( delta_angle > 180 )
			delta_angle -= 360;

		CSunburstCluster* node = this;
		//if a node is the first child of its parent, it can't move leftward;
		while ( node->parent->children == node )
		{
			if ( fabs(node->parent->sweep_angle - 360) < minimum_angle/100 )
			{
				if ( node->siblings == NULL ) //node is the only child;
				{
					node->parent->ChangeStartAngle(new_angle);
					return;
				}
				else
				{
					CSunburstCluster* last_sibling = node->FindLastSibling();
					last_sibling->siblings = node;
					node->parent->start_angle = node->siblings->start_angle;
					node->parent->children = node->siblings;
					node->siblings = NULL;
				}
			}
			else
			{
				node = node->parent;
			}
		}
		
		node->ChangeSweepAngleTowardLeft(delta_angle, minimum_angle);
	}
	else //toward right;
	{
		double delta_angle = new_angle - GetEndAngle();
		if ( delta_angle < -180 )
			delta_angle += 360;
		else if ( delta_angle > 180 )
			delta_angle -= 360;

		CSunburstCluster* node = this;
		//if a node is the last child of its parent, it can't move rightward;
		while ( node->siblings == NULL )
		{
			if ( fabs(node->parent->sweep_angle - 360) < minimum_angle / 100 )
			{
				if ( node->parent->children == node )   //node is the only child 
				{
					double new_start_angle = StandarizeAngle(node->parent->start_angle + delta_angle);
					node->parent->ChangeStartAngle(new_start_angle);
					return;
				}
				else
				{
					CSunburstCluster* before_sibling = node->FindBeforeSibling();
					before_sibling->siblings = NULL;
					node->siblings = node->parent->children;
					node->parent->children = node;
					node->parent->start_angle = node->start_angle;
				}
			}
			else
			{
				node = node->parent;
			}
		}
		
		node->ChangeSweepAngleTowardRight(delta_angle, minimum_angle);

	}
}

//   |<-------------|
//    delta_angle   start_angle
void CSunburstCluster::ChangeSweepAngleTowardLeft(double delta_angle, double minimum_angle)
{
	CSunburstCluster* sibling;

	//change the sweep angles.
    min_angle = minimum_angle;
	double remain_angle = GetAnglesDistance(parent->start_angle, start_angle);

	if ( delta_angle > 0 )  //this sweep angle increases, left sibling's sweep angles decrease
	{
		//calculate alpha for silbings
		if ( delta_angle > remain_angle)
		    alpha = 0;
		else 
			alpha = 1. - delta_angle / remain_angle;

		double total = 0.;

		sibling = parent->children;
		while ( sibling != this )
		{
			sibling->ChangeSweepAngle();
			total += sibling->sweep_angle;
			sibling = sibling->siblings;
		}

		//calculate alpha for this 
		alpha = 1. + ( remain_angle - total ) / sweep_angle;

		this->ChangeSweepAngle();
	}
	else  //this sweep angle decreases, left siblings' sweep angles increase
	{
		double old_sweep_angle = sweep_angle;
		//calculate alpha for this
		if ( sweep_angle + delta_angle < 0 )
		{
			alpha = 0;
		}
		else
		{
			alpha = 1. + delta_angle / sweep_angle; 
		}

		this->ChangeSweepAngle();

		//calculate alpha for siblings
		delta_angle = sweep_angle - old_sweep_angle;
		alpha = 1. - delta_angle / remain_angle;

		sibling = parent->children;
		while ( sibling != this )
		{
			sibling->ChangeSweepAngle();
			sibling = sibling->siblings;
		}
	}

	//change the start angles;
	sibling = parent->children;
	double new_start_angle = sibling->start_angle; 
	while ( sibling != this)
	{
		sibling->ChangeStartAngle(new_start_angle);
		new_start_angle = sibling->GetEndAngle();
		sibling = sibling->siblings;
	}
	this->ChangeStartAngle(new_start_angle);
}



//   |------------->|
//end_angle  delta_angle
void CSunburstCluster::ChangeSweepAngleTowardRight(double delta_angle, double minimum_angle)
{
	CSunburstCluster* sibling;
	//change the sweep angles.
    min_angle = minimum_angle;
    double remain_angle = GetAnglesDistance(GetEndAngle(), parent->GetEndAngle());

	if ( delta_angle > 0 )  //this sweep angle increases, right sibling's sweep angles decrease
	{
		if ( delta_angle > remain_angle)
		    alpha = 0;
		else 
			alpha = 1. - delta_angle / remain_angle;
		
		double total = 0.;

		sibling = this->siblings;
		while ( sibling != NULL )
		{
			sibling->ChangeSweepAngle();
			total += sibling->sweep_angle;
			sibling = sibling->siblings;
		}

		alpha = 1. + ( remain_angle - total ) / sweep_angle;

		this->ChangeSweepAngle();
	}
	else  //this sweep angle decreases, left siblings' sweep angles increase
	{
		double old_sweep_angle = sweep_angle;

		if ( sweep_angle + delta_angle < 0 )
		{
			alpha = 0;
		}
		else
		{
			alpha = 1. + delta_angle / sweep_angle; 
		}

		this->ChangeSweepAngle();

		delta_angle = sweep_angle - old_sweep_angle;
		alpha = 1. - delta_angle / remain_angle;

		sibling = this->siblings;
		while ( sibling != NULL )
		{
			sibling->ChangeSweepAngle();
			sibling = sibling->siblings;
		}
	}

	//change the start angles;
	sibling = this;
	double new_start_angle = sibling->start_angle; 
	while ( sibling != NULL)
	{
		sibling->ChangeStartAngle(new_start_angle);
		new_start_angle = sibling->GetEndAngle();
		sibling = sibling->siblings;
	}
}


//judge if an angle is in the node
//angle: the angle to be judged
//return true is angle is in the node.
bool CSunburstCluster::IsAngleIn(double angle)
{
	if ( angle < start_angle )
		angle += 360;

	if ( angle <= ( start_angle + sweep_angle ) )
		return true;
	else
		return false;
}

//judge if an angle is near the start_angle or end_angle of or in the node. 
//angle: the angle to be judged; eps: the error control parater; 
//return true mean that angle is near the start_angle or end_angle of the node;
//if return true and start = true, angle is near the start_angle of the node;
//if return true and end = true, angle is near the end_angle of the node;
//if return true and in = true, angle is in the end_angle of the node;
bool CSunburstCluster::IsAngleIn(double angle, double eps, bool &start, bool &end, bool &in)
{
    eps = this->sweep_angle / 3.0;
	in = IsAngleIn(angle);
	if ( ! in )
	{
		start = false;
		end = false;
		return false;
	}

	start = IsAngleNear(start_angle, angle, eps);
	end = IsAngleNear(GetEndAngle(), angle, eps);

	if ( start || end || in )
		return true;
	else
		return false;
}


//judge if angle is near angle1
//eps: the error control parater; 
//return true mean that angle is near angle1;
bool CSunburstCluster::IsAngleNear(double angle1, double angle, double eps)
{
	double start_angle1 = angle1 - eps;

	if ( start_angle1 < 0 )
		start_angle1 += 360;

	if ( angle < start_angle1 )
		angle += 360;

	if ( angle <= ( start_angle1 + 2*eps ) )
		return true;
	else
		return false;
}

//Given a level and angle, judge if this node is of this level and covers the angle
CSunburstCluster* CSunburstCluster::FindNodeAccordingLevelAndAngle(int level, double angle)
{
    if ( IsAngleIn(angle) )
    {
        if ( this->level == level )
            return this;
        else if ( this->level < level )
        {
            if ( children != NULL )
                return children->FindNodeAccordingLevelAndAngle(level, angle);
        }
    }
    else if ( siblings != NULL )
    {
        return siblings->FindNodeAccordingLevelAndAngle(level, angle);
    }

    return NULL;
}


//Given a level and angle, judge if this node is of this level, and 
//the angle near the node's start angle
//or near end angle
//or in the angle
//eps: error control parameter;
//start = true means the angle is near the start_angle of the node
//end = true means the angle is near the end_angle of the node
CSunburstCluster* CSunburstCluster::FindNodeAccordingLevelAndAngle(int level, double angle, double eps, 
														 bool &start, bool &end, bool &in)
{
    if ( IsAngleIn(angle, eps, start, end, in) )
    {
        if ( this->level == level )
            return this;
        else if ( this->level < level )
        {
			bool near_this_end = end;
			CSunburstCluster* result = NULL;

            if ( children != NULL )
                result = children->FindNodeAccordingLevelAndAngle(level, angle, eps, start, end, in);
			
			if ( result != NULL || !near_this_end )
				return result;
			else if ( siblings != NULL )
			{
				return siblings->FindNodeAccordingLevelAndAngle(level, angle, eps, start, end, in);
			}
        }
    }
    else if ( siblings != NULL )
    {
        return siblings->FindNodeAccordingLevelAndAngle(level, angle, eps, start, end, in);
    }

    return NULL;
}


int CSunburstCluster::GetEntries()
{
    return entries;
}

double CSunburstCluster::GetRadius()
{
    return radius;
}


void CSunburstCluster::Rotate(double rotate_angle)
{
    this->start_angle += rotate_angle;
    this->StandarizeAngle(this->start_angle);

    if (this->children != NULL )
        this->children->Rotate(rotate_angle);
    if ( this->siblings != NULL )
        this->siblings->Rotate(rotate_angle);
    return;
}
