#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#ifndef _WIN32
	#include <pthread.h>
#else
	#include <windows.h>
#endif

#include "db_buffer.h"
#include "db_loader.h"
#include "db_prefetch.h"
#include "XmdvTool.H"

//jing
#include "CHierNav.H"

#include "db_DISCOVER1.h"
#include "db_GLOBAL.h"

//jing
#include "CHierDisplay.H"

extern int PrefetcherIsRunning;

int	NO_LEAF_NODES;
int	NO_BUFFER_ENTRIES;
ENTRY	**MEMORY = NULL;
DB_ENTRY *TMPDATA = NULL;

int	NO_P_ENTRIES, NO_L_ENTRIES;
T_ENTRY *P = NULL, *L = NULL;

double	old_x = .25, old_y = .75, old_l = 0.;
ENTRY   *current_pointer = NULL;

//jing
#ifndef _WIN32
	pthread_mutex_t mutex;	/* buffer sync mutex */
#else
	HANDLE mutex;	/* buffer sync mutex */
#endif


void destroyBackendConnectivity(void)
{
	if (MEMORY) free(MEMORY);
	if (TMPDATA) free(TMPDATA);
	if (P) free(P);
	if (L) free(L);

	MEMORY = NULL;
	TMPDATA = NULL;
	P = NULL;
	L = NULL;
}

 /*************************************************************************
       error

       error handler
 *************************************************************************/

void error(char *message)
{
	printf("ERROR: %s.\n", message);
	exit(-1);
}


void print_L_table()
{
	int i;
	ENTRY *e;

	puts("L TABLE:");
	for (i=0; i<NO_L_ENTRIES; i++)
	{
		if (!L[i].first) continue;
		printf("(%3d):", i);
		for (e = L[i].first; e; e=e->nextL)
		printf("\t%.2lf %.2lf [%2d]\n", e->data->X, e->data->Y, e->P);
		puts("");
	}
}


 /*************************************************************************
       init_memory

       Initialization routine.
       frees and reallocates MEMORY, TMPDATA, L and P.
       Initializes all the elements of MEMORY and TMPDATA
 *************************************************************************/

void init_memory(char *path)
{
	int index;

	if (!strstr(path, "synt")) return;

    //jing
    #ifndef _WIN32
	    pthread_mutex_init (&mutex, NULL);
	    pthread_mutex_lock (&mutex);
    #else
        mutex = CreateMutex(NULL,          // no security attributes
                            FALSE,         // initially not owned
                            NULL);         // name of mutex
        if (mutex == NULL)                 // error
            return;
        if (WaitForSingleObject(mutex, INFINITE) == WAIT_FAILED)  // failed
            return;
    #endif

	if (MEMORY) free(MEMORY);
	if (TMPDATA) free(TMPDATA);
	if (P) free(P);
	if (L) free(L);

	NO_LEAF_NODES = db_get_leaves(path);
	NO_BUFFER_ENTRIES = (int)(BUFFER_SIZE_RATIO * NO_LEAF_NODES);

	MEMORY = (ENTRY **)calloc(NO_BUFFER_ENTRIES, sizeof(ENTRY *));
	TMPDATA = (DB_ENTRY *)calloc(NO_BUFFER_ENTRIES, sizeof(DB_ENTRY));

	NO_P_ENTRIES = 11; // this might change dynamically in the future
	P = (T_ENTRY *)calloc(NO_P_ENTRIES, sizeof(T_ENTRY));
	
	for (index=0; index<NO_P_ENTRIES; index++)
		init_btable(P + index);

	NO_L_ENTRIES = db_get_levels(path);
	L = (T_ENTRY *)calloc(NO_L_ENTRIES, sizeof(T_ENTRY));
	
	for (index=0; index<NO_L_ENTRIES; index++)
		init_btable(L + index);

	for (index=0; index<NO_BUFFER_ENTRIES; index++)
	{
		MEMORY[index] = new_entry();
		MEMORY[index]->data = new_db_entry();
		P_hash(MEMORY[index]);
	}
	
    //jing
    #ifndef _WIN32
	    pthread_mutex_unlock (&mutex);
    #else
        ReleaseMutex(mutex);
            return;
    #endif
}


 /*************************************************************************
       init_btable

       Initialization routine.
       Used by init_memory to initialize the bucket tables for
       the Probability and the Level table to null.
 *************************************************************************/

void init_btable(T_ENTRY *tablePointer)
{
	tablePointer->first = NULL;
	tablePointer->last  = NULL;
}


 /*************************************************************************
       new_entry

       Creates empty entries for the data to be loaded in the memory.
 *************************************************************************/

ENTRY *new_entry()
{
	ENTRY *entry;

	if ((entry = (ENTRY *)malloc(sizeof(ENTRY))) == NULL)
		error("PERFORMING MALLOC");
	entry->P =  0;
	entry->L = -1;
	entry->data  = NULL;
	entry->nextL = NULL;
	entry->prevL = NULL;
	entry->nextP = NULL;
	entry->prevP = NULL;
	return entry;
}


 /*************************************************************************
       new_db_entry

       Creates empty cluster entries for the data to be loaded in the memory.
 *************************************************************************/

DB_ENTRY *new_db_entry()
{
	int index;
	DB_ENTRY *db_entry;

	if ((db_entry = (DB_ENTRY *)malloc(sizeof(DB_ENTRY))) == NULL)
		error("PERFORMING MALLOC");
	db_entry->X = -1.;
	db_entry->Y = -1.;
	db_entry->L = -1.;
	db_entry->N = -1;

	for (index=1; index<TABLE_ND; index++)
		db_entry->D[index] = 0.;
	return db_entry;
}


 /*************************************************************************
       P_insert_first

       Inserts elements at the start of the Probability table.
 *************************************************************************/

void P_insert_first(T_ENTRY *table, ENTRY *entry)
{
	if (!entry) error("INSERTING A NULL ENTRY");
	entry->prevP = NULL;
	entry->nextP = table->first;
	table->first = entry;
	if (!entry->nextP)
        table->last = entry;
	else
        entry->nextP->prevP = entry;
}


 /*************************************************************************
       P_insert_last

       Inserts elements at the end of the Probability table.
 *************************************************************************/

void P_insert_last(T_ENTRY *table, ENTRY *entry)
{
	if (!entry) error("INSERTING A NULL ENTRY");
	entry->prevP = table->last;
	entry->nextP = NULL;
	table->last = entry;
	if (!entry->prevP)
        table->first = entry;
	else
        entry->prevP->nextP = entry;
}


 /*************************************************************************
       P_hash

       Hashes the elements in the Probability table.
 *************************************************************************/

void P_hash(ENTRY *entry)
{
	if (entry->P == -1)
        error("HASHING AN INVALID ENTRY");  // Invalid entry
	
    P_insert_last(P+entry->P, entry);
}


 /*************************************************************************
       L_insert_first

       Inserts elements at the start of the Level table.
 *************************************************************************/

void L_insert_first(T_ENTRY *table, ENTRY *entry)
{
	if (!entry)
        error("INSERTING A NULL ENTRY");
	entry->prevL = NULL;
	entry->nextL = table->first;
	table->first = entry;
	if (!entry->nextL) 
        table->last = entry;
	else
        entry->nextL->prevL = entry;
}


 /*************************************************************************
       L_insert_last

       Inserts elements at the end of the Level table.
 *************************************************************************/

void L_insert_last(T_ENTRY *table, ENTRY *entry)
{
	if (!entry)
        error("INSERTING A NULL ENTRY");
	entry->prevL = table->last;
	entry->nextL = NULL;
	table->last = entry;
	if (!entry->prevL)
        table->first = entry;
	else
        entry->prevL->nextL = entry;
}


 /*************************************************************************
       L_hash

       Hashes the elements in the Level table.
 *************************************************************************/

void L_hash(ENTRY *entry)
{
	if (entry->L == -1 || !entry->data) error("HASHING AN INVALID ENTRY");

	if (!L[entry->L].first)
		L_insert_first(L+entry->L, entry);
	else if (entry->data->Y <= L[entry->L].first->data->X)
		L_insert_first(L+entry->L, entry);
	else if (entry->data->X >= L[entry->L].last->data->Y)
		L_insert_last(L+entry->L, entry);
	else error("ASSERTION FAILED (hash-1)");
}


 /*************************************************************************
       P_remove_first

       Removes elements from the start of the Probability table.
 *************************************************************************/

ENTRY *P_remove_first(T_ENTRY *table)
{
	ENTRY *first = table->first;

	if (!first)
        return NULL;
    table->first = table->first->nextP;
    if (!table->first)
        table->last = NULL;
	else 
        table->first->prevP = NULL;
    first->prevP = NULL;
	first->nextP = NULL;
	return first;
}


 /*************************************************************************
       P_remove_last

       Removes elements from the end of the Probability table.
 *************************************************************************/

ENTRY *P_remove_last(T_ENTRY *table)
{
	ENTRY *last = table->last;

	if (!last)
        return NULL;
	table->last = table->last->prevP;
	if (!table->last)
        table->first = NULL;
	else
        table->last->nextP = NULL;
	last->prevP = NULL;
	last->nextP = NULL;
	return last;
}


 /*************************************************************************
       P_remove_element

       Removes elements from the Probability table.
 *************************************************************************/

void P_remove_element(ENTRY *entry)
{
	if (!entry->prevP)
	{
		P_remove_first(P+entry->P);
	}
	else if (!entry->nextP)
	{
		P_remove_last(P+entry->P);
	}
	else
	{
		entry->nextP->prevP = entry->prevP;
		entry->prevP->nextP = entry->nextP;
		entry->prevP = NULL;
		entry->nextP = NULL;
	}
}


 /*************************************************************************
       L_remove_first

       Removes elements from the start of the Level table.
 *************************************************************************/

ENTRY *L_remove_first(T_ENTRY *table)
{
	ENTRY *first = table->first;

	if (!first)
        return NULL;
	table->first = table->first->nextL;
	if (!table->first)
        table->last = NULL;
	else
        table->first->prevL = NULL;
	first->prevL = NULL;
	first->nextL = NULL;
	return first;
}


 /*************************************************************************
       L_remove_last

       Removes elements from the end of the Level table.
 *************************************************************************/

ENTRY *L_remove_last(T_ENTRY *table)
{
	ENTRY *last = table->last;

	if (!last)
        return NULL;
	table->last = table->last->prevL;
	if (!table->last)
        table->first = NULL;
	else
        table->last->nextL = NULL;
	last->prevL = NULL;
	last->nextL = NULL;
	return last;
}


 /*************************************************************************
       L_remove_element

       Removes elements from the Level table.
 *************************************************************************/

void L_remove_element(ENTRY *entry)
{
	if (!entry->prevL)
	{
		L_remove_first(P+entry->P);
	}
	else if (!entry->nextL)
	{
		L_remove_last(P+entry->P);
	}
	else
	{
		entry->nextL->prevL = entry->prevL;
		entry->prevL->nextL = entry->nextL;
		entry->prevL = NULL;
		entry->nextL = NULL;
	}
}


 /*     main access routines    */
 /*************************************************************************
       assign_probability

       Assigns probability to the entries passed to it depending
	   on the past exploration by the user and the strategy used
	   for prefetching.
 *************************************************************************/

int assign_probability(ENTRY *e)
{
	double	x = hier_info->tree_brush[0],
		    y = hier_info->tree_brush[1], 
			l = hier_info->brushed_radius;

	int currentLevel = (int) ((NO_L_ENTRIES-1)*l);

	if ((e->L == currentLevel) && (x <= e->data->X) && (e->data->Y <= y))
		return NO_P_ENTRIES-1;

	if ((e->L == currentLevel && e->data->Y <= x && x < old_x) ||
	    (e->L == currentLevel && y <= e->data->X && old_y < y) ||
	    (e->L < currentLevel && old_l > l && x <= e->data->X && e->data->Y <= y) ||
	    (e->L > currentLevel && old_l < l && x <= e->data->X && e->data->Y <= y))
		return (int)((NO_P_ENTRIES-1)*DISC_DIR);

	if (e->L == currentLevel || x <= e->data->X && e->data->Y <= y)
		return Prefetch_Strategy == RANDOM_PREFETCH ? 2 : (int)((NO_P_ENTRIES-1)*((1.0-DISC_DIR)/3.0));

	return 0;
}


 /*************************************************************************
       save_active_set

       Saves the currently active set of the brush int the history.
 *************************************************************************/

void save_active_set()
{
	double	x = hier_info->tree_brush[0], y = hier_info->tree_brush[1], l = hier_info->brushed_radius;

	old_x = x;
	old_y = y;
	old_l = l;
}


 /*************************************************************************
       register_change

       * Recomputes probability for all the entries in the cache
	     and rehashes the table entries.
       * starts the prefetcher.
 *************************************************************************/

void register_change()
{
    double	x = hier_info->tree_brush[0],
			y = hier_info->tree_brush[1],
			l = hier_info->brushed_radius;

	int index, newProbability;
	
	ENTRY *previousEntry, *currentEntry;
	
	if (x==old_x && y==old_y && l==old_l) return;

	stop_prefetcher();

    // recompute probabilities
	for (index=0; index<NO_P_ENTRIES; index++)
		for (currentEntry=P[index].first; currentEntry; )
		{
			previousEntry = currentEntry;
			currentEntry = currentEntry->nextP;
			if (previousEntry->P!=(newProbability=assign_probability(previousEntry)))
			{
				P_remove_element(previousEntry);
				previousEntry->P = newProbability;
				P_hash(previousEntry);
			}
		}

	for (index=0; index<NO_BUFFER_ENTRIES; index++)
		MEMORY[index]->P = assign_probability(MEMORY[index]);

	if (Prefetcher) start_prefetcher();
}


 /*************************************************************************
       load

       Loads data from disk into memory

       The steps are:
       1. test if thread was cancelled	
       2. load data into a temp buffer
       3. select a victim (entry)
       4. remove victim from hash tables
       5. change the content of the entry
       6. rehash entry

 *************************************************************************/

int load(double x, double y, double l, int pi, int first, int last)
{
	int currentLevel = (int) ((NO_L_ENTRIES-1)*l);

	ENTRY *entry;
	int victim,
		index,
		dataDimension,
		totalDataLoaded;
	int over = 0;

	double DELTA = (double)(1<<currentLevel)/(double)NO_LEAF_NODES;
	double XE =  ceil(x/DELTA)*DELTA;
	double YE = floor(y/DELTA)*DELTA;
	double LE = (double)currentLevel/(NO_L_ENTRIES-1);

	fflush(stdout);

//jing
	// test if cancelled
	//pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
	//pthread_testcancel();
	//pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);


	// load data into a temp buffer
    totalDataLoaded = db_load(XE, YE, LE, TMPDATA, last, dataset->datafile);

	for(index=0; index<totalDataLoaded; index++)
	{
//jing
        // test if cancelled
		//pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
		//pthread_testcancel();
		//pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);

		// select a victim
		for (victim=0; victim<NO_P_ENTRIES && !P[victim].first; victim++);
		
		if (victim == NO_P_ENTRIES || pi == NO_P_ENTRIES && victim == NO_P_ENTRIES-1)
			error("ASSERTION FAILED (load-1)");

		if (pi <= P[victim].first->P)
		{
			over = 1;
			break;
		}

		entry = P[victim].first;

		// remove the victim
		P_remove_element(entry);
		if (entry->L == -1);
		else if (entry == L[entry->L].first)
		{
			if (entry->nextL && entry->nextL->nextL)
				P_hash(entry->nextL);
			L_remove_first(L+entry->L);
		}
		else if (entry == L[entry->L].last)
		{
			if (entry->prevL && entry->prevL->prevL)
				P_hash(entry->prevL);
			L_remove_last(L+entry->L);
		}
		else error("ASSERTION FAILED (load-2)");

		// modify entry
		entry->P = pi;
		entry->L = currentLevel;

		entry->data->X = TMPDATA[index].X;
		entry->data->Y = TMPDATA[index].Y;
		entry->data->L = TMPDATA[index].L;
		entry->data->N = 1<<currentLevel;

		for (dataDimension=0; dataDimension<TABLE_ND; dataDimension++)
			entry->data->D[dataDimension] = TMPDATA[index].D[dataDimension];

		// rehash entry
		L_hash(entry);
		
		if (index==0 && last && entry->prevL && entry->prevL->prevL)
			P_remove_element(entry->prevL);
		
		if (index==0 && first && entry->nextL && entry->nextL->nextL)
			P_remove_element(entry->nextL);
		
		if (index==0 && first && last || index==totalDataLoaded-1)
			P_hash(entry);
	}

	if (index && over)	/* if out of space place the last entry in P */
		P_hash(entry);

//jing
	//pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
	return over? -index : index;
}


 /*************************************************************************
       prepare

       Check for data already present in the buffer.
	   Load the new brushed data not present in the buffer.
 *************************************************************************/

int prepare(double x, double y, double l, int pi)
{
	int currentLevel = (int) ((NO_L_ENTRIES-1)*l);
	int over = 0;

    //jing
    #ifndef _WIN32
	    pthread_mutex_lock (&mutex);
    #else
        if (WaitForSingleObject(mutex, INFINITE) == WAIT_FAILED)  // failed
            return 0;
    #endif

    if (L[currentLevel].first && L[currentLevel].last)
	{
		if (x < L[currentLevel].first->data->X &&
			load(x, L[currentLevel].first->data->X, l, pi, 1, 0) < 0)
			over = 1;
		else if (y > L[currentLevel].last->data->Y  &&
			load(L[currentLevel].last->data->Y,  y, l, pi, 0, 1) < 0)
			over = 1;
	}
    else if (load(x, y, l, pi, 1, 1) < 0)
			over = 1;

//jing
	#ifndef _WIN32
		pthread_mutex_unlock (&mutex);
	#else
		if (! ReleaseMutex(mutex))  // failed
        return 0;
	#endif

	return over?-1:0;
}


 /*************************************************************************
       init
	   
       Copy the user selected region to the L link list.
 *************************************************************************/

void init()
{
	double	x = hier_info->tree_brush[0], y = hier_info->tree_brush[1], l = hier_info->brushed_radius;
	int currentLevel = (int) ((NO_L_ENTRIES-1)*l);

	current_pointer = L[currentLevel].first;

	while (	current_pointer && current_pointer->data->X < x)
		current_pointer = current_pointer->nextL;
}


 /*************************************************************************
       get_one
	   
       Returns the current_pointer,
	   set the current_pointer to the next element
 *************************************************************************/

ENTRY *get_one()
{
	double   x = hier_info->tree_brush[0],
		     y = hier_info->tree_brush[1];
    
	ENTRY *aux = current_pointer;

	if (aux && x<=aux->data->X && aux->data->Y<=y)
	{
		current_pointer = aux->nextL;
		return aux;
	}
	else
		return NULL;
}

